/*
 * Copyright (c) 2017, Red Hat Inc. All rights reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Oracle designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Oracle in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */
package org.openjdk.gcbench.fragger;

import org.openjdk.gcbench.util.AllocProfileSupport;
import org.openjdk.gcbench.util.ratelimit.MultiTokenBucket;
import org.openjdk.jmh.annotations.*;

import java.util.LinkedList;
import java.util.List;
import java.util.concurrent.ThreadLocalRandom;
import java.util.concurrent.TimeUnit;

@Warmup(iterations = 5, time = 1, timeUnit = TimeUnit.SECONDS)
@Measurement(iterations = 5, time = 1, timeUnit = TimeUnit.SECONDS)
@Fork(value = 1)
@BenchmarkMode(Mode.Throughput)
@OutputTimeUnit(TimeUnit.SECONDS)
@Threads(Threads.MAX)
@State(Scope.Benchmark)
public class LinkedListFragger {

    public static final int PAYLOAD_SIZE = 16384;

    @Param({"1000"})
    long ldsMB;

    @Param({"10"})
    int rate;

    List<Object> objects;

    MultiTokenBucket bucket;

    int count;

    LinkedList<Object> o0;
    LinkedList<Object> o16;

    private int getSizePerCount() {
        if (AllocProfileSupport.isAvailable()) {
            long a1 = AllocProfileSupport.getAllocatedBytes();
            o0 = new LinkedList<>();
            long a2 = AllocProfileSupport.getAllocatedBytes();
            o16 = new LinkedList<>();
            for (int c = 0; c < 16; c++) {
                o16.add(new byte[PAYLOAD_SIZE]);
            }
            long a3 = AllocProfileSupport.getAllocatedBytes();

            long sizeO0 = a2 - a1;
            long sizeO16 = a3 - a2;

            return (int) ((sizeO16 - sizeO0) / 16);
        } else {
            // best guess
            return 24 + // Node
                    16 + PAYLOAD_SIZE; // payload
        }
    }

    @Setup
    public void setup() {
        bucket = new MultiTokenBucket(rate);
        count = (int)Math.max(1, ldsMB * 1024 * 1024 / getSizePerCount());
        objects = new LinkedList<>();
        for (int c = 0; c < count; c++) {
            objects.add(new byte[PAYLOAD_SIZE]);
        }
    }

    @Benchmark
    public void test() {
        bucket.limit();
        objects.set(ThreadLocalRandom.current().nextInt(count), new byte[PAYLOAD_SIZE]);
    }

}
